@@ -3,8 +3,9 @@ import asyncio
 import json
 from functools import partial
 import uuid
+import ipaddress
 
-from .helpers import get_logger, chown_nobody_permissions
+from .helpers import get_logger, chown_nobody_permissions, PYTHON_GREATER_37
 from .core import Connector
 from .connection import Structures, MessageFields, Misc
 
@@ -14,11 +15,12 @@ DEFAULT_LOGGER_LOG_LEVEL = 'INFO'
 DEFAULT_LOGGER_ROTATE = True
 
 class ConnectorManager:
-    def __init__(self, config_file_path=None, logger=None, use_default_logger=True, 
+    def __init__(self, config_file_path=None, config_file_overrides_kwargs=True, logger=None, use_default_logger=True, 
                  default_logger_log_level=DEFAULT_LOGGER_LOG_LEVEL, default_logger_rotate=True,
-                 default_logger_dirpath=Connector.CONNECTOR_FILES_DIRPATH,
-                 is_server=True, server_sockaddr=None, use_ssl=Connector.USE_SSL, ssl_allow_all=False, 
-                 client_bind_ip=None, certificates_directory_path=None, client_name=None, 
+                 default_logger_dirpath=Connector.CONNECTOR_FILES_DIRPATH, use_token=Connector.USE_TOKEN,
+                 is_server=True, server_sockaddr=None, use_ssl=Connector.USE_SSL, ssl_allow_all=Connector.SSL_ALLOW_ALL, 
+                 server_ca=Connector.SERVER_CA, client_bind_ip=None, certificates_directory_path=Connector.CONNECTOR_FILES_DIRPATH,
+                 tokens_directory_path=Connector.CONNECTOR_FILES_DIRPATH, client_name=None, 
                  send_message_types=None, recv_message_types=None, subscribe_message_types=None,
                  connector_files_dirpath=Connector.CONNECTOR_FILES_DIRPATH,
                  disk_persistence_send=Connector.DISK_PERSISTENCE_SEND, 
@@ -28,11 +30,21 @@ class ConnectorManager:
                  uds_path_receive_preserve_socket=Connector.UDS_PATH_RECEIVE_PRESERVE_SOCKET,
                  uds_path_send_preserve_socket=Connector.UDS_PATH_SEND_PRESERVE_SOCKET,
                  hook_server_auth_client=None, hook_target_directory=None, enable_client_try_reconnect=True,
+                 keep_alive_period=None, keep_alive_timeout=Connector.KEEP_ALIVE_TIMEOUT, send_timeout=Connector.SEND_TIMEOUT,
+                 max_number_of_unanswered_keep_alive=Connector.MAX_NUMBER_OF_UNANSWERED_KEEP_ALIVE,
                  reuse_server_sockaddr=False, reuse_uds_path_send_to_connector=False, reuse_uds_path_commander_server=False,
-                 max_size_file_upload=None, hook_allow_certificate_creation=None,
+                 max_size_file_upload=None, hook_allow_certificate_creation=None, hook_proxy_authorization=None,
+                 hook_store_token=None, hook_load_token=None,
                  max_size_file_upload_send=Connector.MAX_SIZE_FILE_UPLOAD_SEND, max_size_file_upload_recv=Connector.MAX_SIZE_FILE_UPLOAD_RECV,
                  everybody_can_send_messages=Connector.EVERYBODY_CAN_SEND_MESSAGES, max_certs=Connector.MAX_CERTS,
-                 send_message_types_priorities=None, pubsub_central_broker=False, proxy=None):
+                 send_message_types_priorities=None, pubsub_central_broker=False, proxy=None,
+                 alternate_client_default_cert=Connector.ALTERNATE_CLIENT_DEFAULT_CERT,
+                 blacklisted_clients_id=None, blacklisted_clients_ip=None, blacklisted_clients_subnet=None,
+                 whitelisted_clients_id=None, whitelisted_clients_ip=None, whitelisted_clients_subnet=None,
+                 hook_whitelist_clients=None, ignore_peer_traffic=False,
+                 token_verify_peer_cert=Connector.TOKEN_VERIFY_PEER_CERT, token_client_send_cert=Connector.TOKEN_CLIENT_SEND_CERT,
+                 token_client_verify_server_hostname=Connector.TOKEN_CLIENT_VERIFY_SERVER_HOSTNAME,
+                 token_server_allow_authorized_non_default_cert=False):
         
         self.connector_files_dirpath = connector_files_dirpath
         self.default_logger_dirpath = default_logger_dirpath
@@ -53,9 +65,13 @@ class ConnectorManager:
         else:
             self.logger = logger
             
-        self.is_server, self.server_sockaddr, self.use_ssl, self.ssl_allow_all, self.certificates_directory_path = \
-                            is_server, server_sockaddr, use_ssl, ssl_allow_all, certificates_directory_path
-        self.client_name, self.client_bind_ip = client_name, client_bind_ip
+        self.is_server, self.server_sockaddr, self.use_ssl, self.ssl_allow_all, self.certificates_directory_path, self.server_ca = \
+                            is_server, server_sockaddr, use_ssl, ssl_allow_all, certificates_directory_path, server_ca
+        self.client_name, self.client_bind_ip, self.tokens_directory_path, self.use_token = client_name, client_bind_ip,\
+                            tokens_directory_path, use_token
+        self.token_verify_peer_cert, self.token_client_send_cert = token_verify_peer_cert, token_client_send_cert
+        self.token_client_verify_server_hostname = token_client_verify_server_hostname
+        self.token_server_allow_authorized_non_default_cert = token_server_allow_authorized_non_default_cert
         self.send_message_types, self.recv_message_types = send_message_types, recv_message_types
         self.pubsub_central_broker = pubsub_central_broker        
         self.subscribe_message_types = subscribe_message_types
@@ -68,6 +84,7 @@ class ConnectorManager:
         self.send_message_types_priorities = send_message_types_priorities
         self.proxy = proxy
         self.max_certs = max_certs
+        self.alternate_client_default_cert = alternate_client_default_cert
 
         self.disk_persistence_send, self.disk_persistence_recv, self.max_size_persistence_path = \
                             disk_persistence_send, disk_persistence_recv, max_size_persistence_path
@@ -76,12 +93,22 @@ class ConnectorManager:
                             uds_path_receive_preserve_socket, uds_path_send_preserve_socket
         self.hook_server_auth_client, self.hook_target_directory = hook_server_auth_client, hook_target_directory
         self.hook_allow_certificate_creation = hook_allow_certificate_creation
+        self.hook_proxy_authorization = hook_proxy_authorization
+        self.hook_store_token, self.hook_load_token = hook_store_token, hook_load_token
         self.enable_client_try_reconnect = enable_client_try_reconnect
         self.reuse_server_sockaddr, self.reuse_uds_path_send_to_connector, self.reuse_uds_path_commander_server = \
                             reuse_server_sockaddr, reuse_uds_path_send_to_connector, reuse_uds_path_commander_server
-
-        self.config_file_path = config_file_path
-        if self.config_file_path:
+        self.keep_alive_period, self.keep_alive_timeout, self.max_number_of_unanswered_keep_alive = \
+                    keep_alive_period, keep_alive_timeout, max_number_of_unanswered_keep_alive
+        self.send_timeout = send_timeout
+        
+        self.blacklisted_clients_id, self.blacklisted_clients_ip = blacklisted_clients_id, blacklisted_clients_ip
+        self.blacklisted_clients_subnet, self.whitelisted_clients_id = blacklisted_clients_subnet, whitelisted_clients_id
+        self.whitelisted_clients_ip, self.whitelisted_clients_subnet = whitelisted_clients_ip, whitelisted_clients_subnet
+        self.hook_whitelist_clients, self.ignore_peer_traffic = hook_whitelist_clients, ignore_peer_traffic
+        
+        self.config_file_path, self.config_file_overrides_kwargs = config_file_path, config_file_overrides_kwargs
+        if self.config_file_path and self.config_file_overrides_kwargs:
             self.config_file_path = str(self.config_file_path)
             if os.path.exists(self.config_file_path):
                 try:
@@ -129,15 +156,16 @@ class ConnectorManager:
         self.send_message_types.append('_pubsub')        
         
         self.connector = Connector(self.logger, is_server=self.is_server, server_sockaddr=self.server_sockaddr, 
-                                   use_ssl=self.use_ssl, ssl_allow_all=self.ssl_allow_all,
-                                   certificates_directory_path=self.certificates_directory_path, 
+                                   use_ssl=self.use_ssl, ssl_allow_all=self.ssl_allow_all, use_token=self.use_token,
+                                   server_ca=self.server_ca, certificates_directory_path=self.certificates_directory_path,
+                                   tokens_directory_path=self.tokens_directory_path,
                                    client_name=self.client_name, client_bind_ip=self.client_bind_ip,
                                    send_message_types=self.send_message_types, recv_message_types=self.recv_message_types,
                                    subscribe_message_types=self.subscribe_message_types,
                                    disk_persistence_send=self.disk_persistence_send, 
                                    disk_persistence_recv=self.disk_persistence_recv,
                                    max_size_persistence_path=self.max_size_persistence_path, 
-                                   file_recv_config=self.file_recv_config,
+                                   file_recv_config=self.file_recv_config, config_file_path=self.config_file_path,
                                    debug_msg_counts=self.debug_msg_counts, silent=self.silent, 
                                    connector_files_dirpath=self.connector_files_dirpath, #use_ack=use_ack,
                                    uds_path_receive_preserve_socket=self.uds_path_receive_preserve_socket, 
@@ -145,8 +173,12 @@ class ConnectorManager:
                                    hook_server_auth_client=self.hook_server_auth_client,
                                    hook_target_directory=self.hook_target_directory,
                                    hook_allow_certificate_creation=self.hook_allow_certificate_creation,
+                                   hook_proxy_authorization=self.hook_proxy_authorization,
+                                   hook_store_token=self.hook_store_token, hook_load_token=self.hook_load_token,
                                    enable_client_try_reconnect=self.enable_client_try_reconnect,
-                                   reuse_server_sockaddr=self.reuse_server_sockaddr,
+                                   keep_alive_period=self.keep_alive_period, keep_alive_timeout=self.keep_alive_timeout,
+                                   max_number_of_unanswered_keep_alive=self.max_number_of_unanswered_keep_alive,
+                                   reuse_server_sockaddr=self.reuse_server_sockaddr, send_timeout = self.send_timeout,
                                    reuse_uds_path_send_to_connector=self.reuse_uds_path_send_to_connector,
                                    reuse_uds_path_commander_server=self.reuse_uds_path_commander_server,
                                    max_size_file_upload_send=self.max_size_file_upload_send,
@@ -154,7 +186,15 @@ class ConnectorManager:
                                    everybody_can_send_messages=self.everybody_can_send_messages,
                                    send_message_types_priorities=self.send_message_types_priorities,
                                    pubsub_central_broker=self.pubsub_central_broker, proxy=self.proxy,
-                                   max_certs=self.max_certs)        
+                                   max_certs=self.max_certs, alternate_client_default_cert=self.alternate_client_default_cert,
+                                   blacklisted_clients_id=self.blacklisted_clients_id, blacklisted_clients_ip=self.blacklisted_clients_ip,
+                                   blacklisted_clients_subnet=self.blacklisted_clients_subnet, whitelisted_clients_id=self.whitelisted_clients_id,
+                                   whitelisted_clients_ip=self.whitelisted_clients_ip, whitelisted_clients_subnet=self.whitelisted_clients_subnet,
+                                   hook_whitelist_clients=self.hook_whitelist_clients,
+                                   token_verify_peer_cert=self.token_verify_peer_cert, token_client_send_cert=self.token_client_send_cert,
+                                   token_client_verify_server_hostname=self.token_client_verify_server_hostname,
+                                   token_server_allow_authorized_non_default_cert=self.token_server_allow_authorized_non_default_cert
+)
         
             
     async def start_connector(self, delay=None, connector_socket_only=False):        
@@ -246,6 +286,38 @@ class ConnectorManager:
         else:
             res = False
         return res
+
+    async def add_blacklist_client(self, client_ip=None, client_id=None):
+        self.logger.info(f'{self.source_id} blacklist_client ip : {client_ip}, id : {client_ip}')
+        if self.connector.is_server:
+            response = await self.connector.add_blacklist_client(client_ip=client_ip, client_id=client_id)
+            return response
+        else:
+            return False
+
+    async def remove_blacklist_client(self, client_ip=None, client_id=None):
+        self.logger.info(f'{self.source_id} remove_blacklist_client ip : {client_ip}, id : {client_ip}')
+        if self.connector.is_server:
+            response = await self.connector.remove_blacklist_client(client_ip=client_ip, client_id=client_id)
+            return response
+        else:
+            return False
+    
+    async def add_whitelist_client(self, client_ip=None, client_id=None):
+        self.logger.info(f'{self.source_id} whitelist_client ip : {client_ip}, id : {client_ip}')
+        if self.connector.is_server:
+            response = await self.connector.add_whitelist_client(client_ip=client_ip, client_id=client_id)
+            return response
+        else:
+            return False
+
+    async def remove_whitelist_client(self, client_ip=None, client_id=None):
+        self.logger.info(f'{self.source_id} remove_whitelist_client ip : {client_ip}, id : {client_ip}')
+        if self.connector.is_server:
+            response = await self.connector.remove_whitelist_client(client_ip=client_ip, client_id=client_id)
+            return response
+        else:
+            return False
     
     async def delete_client_certificate(self, client_id, remove_only_symlink=False):
         self.logger.info(f'{self.source_id} delete_client_certificate {client_id} with'
@@ -256,6 +328,22 @@ class ConnectorManager:
             res = await self.connector.delete_client_certificate_on_client()
         return res     
 
+    async def delete_client_token(self, client_id):
+        self.logger.info(f'{self.source_id} delete_client_token {client_id}')
+        if self.connector.is_server:
+            res = await self.connector.delete_client_token_on_server(client_id)
+        else:
+            res = await self.connector.delete_client_token_on_client()
+        return res     
+
+    async def reload_tokens(self):
+        self.logger.info(f'{self.source_id} reload_tokens')
+        if self.connector.is_server:
+            res = await self.connector.load_server_tokens()
+        else:
+            res = await self.connector.load_client_token()
+        return res     
+
     def show_subscribe_message_types(self):
         self.logger.info(f'{self.source_id} show_subscribe_message_types')
         return self.connector.show_subscribe_message_types()
@@ -542,6 +630,12 @@ class ConnectorAPI(ConnectorBaseTool):
                         self.logger.exception('send_message uds_path_send_preserve_socket')
                         try:
                             writer.close()
+                            if PYTHON_GREATER_37:
+                                try:
+                                    await writer.wait_closed()      #python 3.7                                              
+                                except Exception as exc:
+                                    self.logger.warning('send_message1 wait_closed : '+str(exc))
+                            
                         except Exception:
                             pass
                                           
@@ -557,6 +651,7 @@ class ConnectorAPI(ConnectorBaseTool):
                 
                 reader, writer = await asyncio.wait_for(asyncio.open_unix_connection(path=self.connector.uds_path_send_to_connector, 
                                                    limit=Connector.MAX_SOCKET_BUFFER_SIZE), timeout=Connector.ASYNC_TIMEOUT)
+                writer.transport.set_write_buffer_limits(0,0)
                 if self.uds_path_send_preserve_socket and not await_response:
                     self.reader_writer_uds_path_send = reader, writer
             except asyncio.CancelledError:
@@ -584,15 +679,29 @@ class ConnectorAPI(ConnectorBaseTool):
                 if await_response_timeout is not None:
                     try:
                         the_response = await asyncio.wait_for(self.recv_message(reader, writer), timeout=await_response_timeout)
+                        if the_response[0].get(MessageFields.ERROR):
+                            raise asyncio.IncompleteReadError(the_response[0].get(MessageFields.ERROR).encode(), 0)
                     except asyncio.TimeoutError:
                         self.logger.warning(f'send_message : await_response_timeout error ({await_response_timeout} s)')
                         writer.close()
+                        if PYTHON_GREATER_37:
+                            try:
+                                await writer.wait_closed()      #python 3.7                                              
+                            except Exception as exc:
+                                self.logger.warning('send_message2 wait_closed : '+str(exc))
+                        
                         return False                      
                 else:
                     the_response = await self.recv_message(reader, writer)
             self.logger.debug('send_message finished sending')                    
             if await_response:
                 writer.close()
+                if PYTHON_GREATER_37:
+                    try:
+                        await writer.wait_closed()      #python 3.7                                              
+                    except Exception as exc:
+                        self.logger.warning('send_message3 wait_closed : '+str(exc))
+                
                 return the_response
             return True      
                 
@@ -604,7 +713,7 @@ class ConnectorAPI(ConnectorBaseTool):
             return False
         except ConnectionResetError as exc:
             self.logger.warning('ConnectionResetError : '+str(exc))
-        except Exception as exc:
+        except Exception:
             self.logger.exception('send_data')
             return False
 
@@ -640,12 +749,12 @@ class ConnectorAPI(ConnectorBaseTool):
         except asyncio.CancelledError:
             raise        
         except asyncio.IncompleteReadError:
-            self.logger.warning('recv_message : peer disconnected')
+            self.logger.info('recv_message : peer disconnected')
             return None, None, None
         except ConnectionResetError as exc:
             self.logger.warning('recv_message : peer disconnected '+str(exc))
             return None, None, None            
-        except Exception as exc:
+        except Exception:
             self.logger.exception('recv_message')
             raise            
 
@@ -689,11 +798,11 @@ class ConnectorAPI(ConnectorBaseTool):
                                                      limit=Connector.MAX_SOCKET_BUFFER_SIZE)
             self.message_waiters[message_type] = server
             if self.receive_from_any_connector_owner:
-                chown_nobody_permissions(uds_path_receive_from_connector, self.logger)            
+                chown_nobody_permissions(uds_path_receive_from_connector, self.logger)   
             return server
         except asyncio.CancelledError:
             raise        
-        except Exception as exc:
+        except Exception:
             self.logger.exception('start_waiting_for_messages')
             raise
 
@@ -728,6 +837,7 @@ class ConnectorRemoteTool(ConnectorBaseTool):
             message = Structures.MSG_4_STRUCT.pack(len(message)) + message
             reader, writer = await asyncio.wait_for(asyncio.open_unix_connection(path=self.connector.uds_path_commander), 
                                                     timeout=Connector.ASYNC_TIMEOUT)
+            writer.transport.set_write_buffer_limits(0,0)
             writer.write(message)  
             try:
                 await asyncio.wait_for(writer.drain(), timeout=Connector.ASYNC_TIMEOUT)
@@ -737,6 +847,12 @@ class ConnectorRemoteTool(ConnectorBaseTool):
             next_length = Structures.MSG_4_STRUCT.unpack(next_length_bytes)[0]
             response = await asyncio.wait_for(reader.readexactly(next_length), timeout=Connector.ASYNC_TIMEOUT)
             writer.close()
+            if PYTHON_GREATER_37:
+                try:
+                    await writer.wait_closed()      #python 3.7                                              
+                except Exception as exc:
+                    self.logger.warning('send_command wait_closed : '+str(exc))
+            
             self.logger.info(f'send_command got response {response}')
         except Exception as exc:
             self.logger.exception('send_command')
@@ -780,6 +896,25 @@ class ConnectorRemoteTool(ConnectorBaseTool):
         else:
             response = await self.send_command(cmd='delete_client_certificate_on_client', kwargs={'restart_client':restart_client})
             return response
+
+    async def delete_client_token(self, client_id=None, restart_client=True):
+        self.logger.info(f'{self.source_id} delete_client_token {client_id}')                 
+        if self.is_server:
+            response = await self.send_command(cmd='delete_client_token_on_server', 
+                                               kwargs={'client_id':client_id})
+            return response
+        else:
+            response = await self.send_command(cmd='delete_client_token_on_client', kwargs={'restart_client':restart_client})
+            return response
+
+    async def reload_tokens(self):
+        self.logger.info(f'{self.source_id} reload_tokens')                 
+        if self.is_server:
+            response = await self.send_command(cmd='load_server_tokens')
+            return response
+        else:
+            response = await self.send_command(cmd='load_client_token')
+            return response
         
     async def disconnect_client(self, client_id=None):
         self.logger.info(f'{self.source_id} disconnect_client {client_id}')                         
@@ -788,6 +923,38 @@ class ConnectorRemoteTool(ConnectorBaseTool):
             return response
         else:
             return False
+
+    async def add_blacklist_client(self, client_ip=None, client_id=None):
+        self.logger.info(f'{self.source_id} blacklist_client ip : {client_ip}, id : {client_ip}')
+        if self.is_server:
+            response = await self.send_command(cmd='add_blacklist_client', kwargs={'client_ip':client_ip, 'client_id':client_id})
+            return response
+        else:
+            return False
+
+    async def remove_blacklist_client(self, client_ip=None, client_id=None):
+        self.logger.info(f'{self.source_id} remove_blacklist_client ip : {client_ip}, id : {client_ip}')
+        if self.is_server:
+            response = await self.send_command(cmd='remove_blacklist_client', kwargs={'client_ip':client_ip, 'client_id':client_id})
+            return response
+        else:
+            return False
+
+    async def add_whitelist_client(self, client_ip=None, client_id=None):
+        self.logger.info(f'{self.source_id} whitelist_client ip : {client_ip}, id : {client_ip}')
+        if self.is_server:      
+            response = await self.send_command(cmd='add_whitelist_client', kwargs={'client_ip':client_ip, 'client_id':client_id})
+            return response
+        else:
+            return False
+
+    async def remove_whitelist_client(self, client_ip=None, client_id=None):
+        self.logger.info(f'{self.source_id} remove_whitelist_client ip : {client_ip}, id : {client_ip}')
+        if self.is_server:
+            response = await self.send_command(cmd='remove_whitelist_client', kwargs={'client_ip':client_ip, 'client_id':client_id})
+            return response
+        else:
+            return False
         
     async def delete_previous_persistence_remains(self):
         self.logger.info(f'{self.source_id} delete_previous_persistence_remains')         
@@ -804,9 +971,9 @@ class ConnectorRemoteTool(ConnectorBaseTool):
         response = await self.send_command(cmd='set_subscribe_message_types', kwargs={'message_types':message_types})
         return response
     
-    async def show_connected_peers(self):
+    async def show_connected_peers(self, dump_result=True):
         self.logger.info(f'{self.source_id} show_connected_peers')         
-        response = await self.send_command(cmd='show_connected_peers__sync', kwargs={})        
+        response = await self.send_command(cmd='show_connected_peers__sync', kwargs={'dump_result':dump_result})        
         return response
 
     async def peek_queues(self):
@@ -820,17 +987,17 @@ class ConnectorRemoteTool(ConnectorBaseTool):
         return response
 
     async def ignore_peer_traffic_enable(self):
-        self.logger.info(f'{self.source_id} ignore_peer_traffic_enable')         
+        self.logger.info(f'{self.source_id} ignore_peer_traffic_enable')
         response = await self.send_command(cmd='manage_ignore_peer_traffic__sync', kwargs={'enable':True})        
         return response
     
     async def ignore_peer_traffic_enable_unique(self, peername):
-        self.logger.info(f'{self.source_id} ignore_peer_traffic_enable_unique')         
+        self.logger.info(f'{self.source_id} ignore_peer_traffic_enable_unique')
         response = await self.send_command(cmd='manage_ignore_peer_traffic__sync', kwargs={'unique_peer':peername})        
         return response
 
     async def ignore_peer_traffic_disable(self):
-        self.logger.info(f'{self.source_id} ignore_peer_traffic_disable')         
+        self.logger.info(f'{self.source_id} ignore_peer_traffic_disable')
         response = await self.send_command(cmd='manage_ignore_peer_traffic__sync', kwargs={'disable':True})        
         return response    
 
@@ -843,11 +1010,10 @@ class ConnectorRemoteTool(ConnectorBaseTool):
         self.logger.info(f'{self.source_id} set_log_level {level}')         
         response = await self.send_command(cmd='set_log_level__sync', kwargs={'level':level})        
         return response
-    
-      
 
-            
-            
-            
-            
-            
+    async def show_attribute(self, attribute):
+        self.logger.info(f'{self.source_id} show_attribute {attribute}')         
+        response = await self.send_command(cmd='show_attribute__sync', kwargs={'attribute':attribute})        
+        return response
+    
+    
