@@ -11,6 +11,8 @@ import re
 
 import aioconnectors
 
+REGEX_IP = re.compile('^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}')
+
 def create_connector(config_file_path, logger=None):
     if not logger:
         logger = aioconnectors.get_logger(logger_name='create_connector', first_run=True)    
@@ -45,7 +47,16 @@ def cli(logger=None):
             print(str(the_key)+')    '+str(value))
     
     if not logger:
-        logger = aioconnectors.get_logger(logger_name='cli', first_run=True)    
+        logger = aioconnectors.get_logger(logger_name='cli', first_run=True)
+    running_with_tab_completion = True
+    try:    
+        import readline
+        readline.set_completer_delims('\t\n= ')
+        readline.parse_and_bind('tab:complete')
+    except Exception:
+        running_with_tab_completion = False
+        logger.info('Running without tab completion')
+        
     print('\nWelcome to aioconnectors CLI')
     Connector = aioconnectors.core.Connector    
     the_path = input('\nPlease type your connector_files_dirpath, or Enter if it is '
@@ -54,7 +65,7 @@ def cli(logger=None):
         the_path = os.path.abspath(os.path.normpath(os.path.expandvars(os.path.expanduser(the_path))))
     else:
         the_path = Connector.CONNECTOR_FILES_DIRPATH
-    
+        
     while True:
         active_connectors_path = os.path.join(the_path, Connector.DEFAULT_ACTIVE_CONNECTORS_NAME)
         dict_connector_names = {}
@@ -98,15 +109,27 @@ def cli(logger=None):
             continue
         clearscreen()
         list_cmds = ['start', 'stop gracefully', 'stop hard', 'restart', 'show_connected_peers', 
-                     'ignore_peer_traffic', 'peek_queues', 'delete_client_certificate', 'disconnect_client',
-                     'show_log_level', 'set_log_level', 'show_subscribe_message_types', 'set_subscribe_message_types']
+                     'ignore_peer_traffic', 'peek_queues', 'delete_client_certificate', 'delete_client_token', 'disconnect_client',
+                     'show_log_level', 'set_log_level', 'show_subscribe_message_types', 'set_subscribe_message_types',
+                     'add_blacklist_client', 'remove_blacklist_client', 'add_whitelist_client', 'remove_whitelist_client']
         dict_cmds = {str(index):cmd for index,cmd in enumerate(list_cmds)}
         display_dict(dict_cmds, connector=server_sockaddr or client_name)        
         res = input('\nPlease type the command number you would like to run, or q to quit\n')
-        def show_connected_peers():
+
+        def show_connected_peers(return_peers=False):
             task = loop.create_task(connector_remote_tool.show_connected_peers())
             loop.run_until_complete(task)
-            print(f'\nConnected peers : {task.result().decode()}')
+            peers_dict = task.result().decode()
+            print(f'\nConnected peers : {peers_dict}')
+            if return_peers:
+                return json.loads(peers_dict)
+            
+        def show_attribute(attribute):
+            show_attribute            
+            task = loop.create_task(connector_remote_tool.show_attribute(attribute))
+            loop.run_until_complete(task)
+            value = task.result().decode()
+            print(f'\n{attribute.capitalize()} : {value}')
             
         while True:
             clearscreen()
@@ -116,24 +139,30 @@ def cli(logger=None):
                 print('Invalid number : '+str(res))
             else:
                 the_cmd = dict_cmds[res]
+                
                 if the_cmd == 'start':
                     task = loop.create_task(connector_remote_tool.start_connector())
                     loop.run_until_complete(task)
-                    print(task.result().decode())                        
+                    print(task.result().decode())           
+                    
                 elif the_cmd == 'stop gracefully':
                     task = loop.create_task(connector_remote_tool.stop_connector(client_wait_for_reconnect=False, hard=False))
                     loop.run_until_complete(task)
                     print(task.result().decode())
+                    
                 elif the_cmd == 'stop hard':
                     task = loop.create_task(connector_remote_tool.stop_connector(client_wait_for_reconnect=False, hard=True))
                     loop.run_until_complete(task)
-                    print(task.result().decode())                    
+                    print(task.result().decode())          
+                    
                 elif the_cmd == 'restart':
                     task = loop.create_task(connector_remote_tool.restart_connector(sleep_between=2, hard=False))
                     loop.run_until_complete(task)
                     print(task.result().decode())
+                    
                 elif the_cmd == 'show_connected_peers':
-                    show_connected_peers()                     
+                    show_connected_peers()             
+                    
                 elif the_cmd == 'ignore_peer_traffic':
                     while True:
                         task = loop.create_task(connector_remote_tool.ignore_peer_traffic_show())
@@ -141,14 +170,25 @@ def cli(logger=None):
                         status = task.result().decode()
                         print('\nignore_peer_traffic current status : ', status)
                         if status == 'False':
-                            show_connected_peers()
+                            
+                            peers_dict = show_connected_peers(return_peers=running_with_tab_completion)
+                            if running_with_tab_completion:
+                                def complete(text,state):
+                                    results = [peer for peer in peers_dict if peer.startswith(text)] + [None]
+                                    return results[state]
+                                readline.set_completer(complete)                        
+
                             res = input('\nType "y" to ignore peer traffic, or <peer name> to ignore a unique peer '
-                                        'traffic, or Enter to quit\n')
+                                        'traffic, or q to quit\n')
+                            
+                            if running_with_tab_completion:
+                                readline.set_completer(None) 
+                            
                             if res == 'y':
                                 task = loop.create_task(connector_remote_tool.ignore_peer_traffic_enable())
                                 loop.run_until_complete(task)
                                 continue
-                            elif res == '':
+                            elif res == 'q':
                                 break
                             else:
                                 task = loop.create_task(connector_remote_tool.ignore_peer_traffic_enable_unique(res))
@@ -161,47 +201,221 @@ def cli(logger=None):
                                 loop.run_until_complete(task)
                                 continue
                             else:
-                                break                                
+                                break       
+                            
                 elif the_cmd == 'peek_queues':
                     task = loop.create_task(connector_remote_tool.peek_queues())
                     loop.run_until_complete(task)
                     print(json.dumps(json.loads(task.result().decode()), indent=4, sort_keys=True))
+                    
                 elif the_cmd == 'delete_client_certificate':
                     if is_server:
-                        show_connected_peers()
+                        peers_dict = show_connected_peers(return_peers=running_with_tab_completion)
+                        if running_with_tab_completion:
+                            def complete(text,state):
+                                results = [peer for peer in peers_dict if peer.startswith(text)] + [None]
+                                return results[state]
+                            readline.set_completer(complete)                        
+                        
                         client_name = input('\nPlease type the client name whose certificate you would '
                                             'like to delete, or q to quit\n')
-                        res = input('\nAre you sure you want to delete '+client_name+' \'s certificate ? y/n\n')
-                        if res =='y':
-                            task = loop.create_task(connector_remote_tool.delete_client_certificate(client_id=client_name, 
-                                                                                            remove_only_symlink=False))
-                            loop.run_until_complete(task)
-                            print(task.result().decode())                            
-                            task = loop.create_task(connector_remote_tool.disconnect_client(client_id=client_name))
-                            loop.run_until_complete(task)
-                            print(task.result().decode())                            
+                        if running_with_tab_completion:
+                            readline.set_completer(None) 
+                        
+                        if client_name != 'q':
+                            res = input('\nAre you sure you want to delete '+client_name+' \'s certificate ? y/n\n')
+                            if res =='y':
+                                task = loop.create_task(connector_remote_tool.delete_client_certificate(client_id=client_name, 
+                                                                                                remove_only_symlink=False))
+                                loop.run_until_complete(task)
+                                print(task.result().decode())                            
+                                task = loop.create_task(connector_remote_tool.disconnect_client(client_id=client_name))
+                                loop.run_until_complete(task)
+                                print(task.result().decode())                            
                     else:
                         res = input('\nAre you sure you want to delete '+client_name+' \'s certificate ? y/n\n')
                         if res =='y':                            
                             task = loop.create_task(connector_remote_tool.delete_client_certificate())
                             loop.run_until_complete(task)
                             print(task.result().decode())
+
+                elif the_cmd == 'delete_client_token':
+                    if is_server:
+                        peers_dict = show_connected_peers(return_peers=running_with_tab_completion)
+                        if running_with_tab_completion:
+                            def complete(text,state):
+                                results = [peer for peer in peers_dict if peer.startswith(text)] + [None]
+                                return results[state]
+                            readline.set_completer(complete)                        
+                        
+                        client_name = input('\nPlease type the client name whose token you would '
+                                            'like to delete, or q to quit\n')
+                        if running_with_tab_completion:
+                            readline.set_completer(None) 
+                        
+                        if client_name != 'q':
+                            res = input('\nAre you sure you want to delete '+client_name+' \'s token ? y/n\n')
+                            if res =='y':
+                                task = loop.create_task(connector_remote_tool.delete_client_token(client_id=client_name))
+                                loop.run_until_complete(task)
+                                print(task.result().decode())                            
+                                task = loop.create_task(connector_remote_tool.disconnect_client(client_id=client_name))
+                                loop.run_until_complete(task)
+                                print(task.result().decode())                            
+                    else:
+                        res = input('\nAre you sure you want to delete '+client_name+' \'s token ? y/n\n')
+                        if res =='y':                            
+                            task = loop.create_task(connector_remote_tool.delete_client_token())
+                            loop.run_until_complete(task)
+                            print(task.result().decode())
+                            
                 elif the_cmd == 'disconnect_client':
                     if is_server:
-                        show_connected_peers()
+                        peers_dict = show_connected_peers(return_peers=running_with_tab_completion)
+                        if running_with_tab_completion:
+                            def complete(text,state):
+                                results = [peer for peer in peers_dict if peer.startswith(text)] + [None]
+                                return results[state]
+                            readline.set_completer(complete)                        
+                        
                         client_name = input('\nPlease type the client name you would '
                                             'like to disconnect, or q to quit\n')
-                        res = input('\nAre you sure you want to disconnect '+client_name+' ? y/n\n')
-                        if res =='y':
-                            task = loop.create_task(connector_remote_tool.disconnect_client(client_id=client_name))
-                            loop.run_until_complete(task)
-                            print(task.result().decode())                          
+                        if running_with_tab_completion:
+                            readline.set_completer(None) 
+                            
+                        if client_name != 'q':                        
+                            res = input('\nAre you sure you want to disconnect '+client_name+' ? y/n\n')
+                            if res =='y':
+                                task = loop.create_task(connector_remote_tool.disconnect_client(client_id=client_name))
+                                loop.run_until_complete(task)
+                                print(task.result().decode())                          
                     else:
-                        print('A client cannot use this functionality')                        
+                        print('A client cannot use this functionality')         
+
+                elif the_cmd == 'add_blacklist_client':
+                    if is_server:
+                        show_attribute('blacklisted_clients_id')
+                        show_attribute('blacklisted_clients_ip')
+                        show_attribute('blacklisted_clients_subnet')                        
+                        peers_dict = show_connected_peers(return_peers=running_with_tab_completion)
+                        if running_with_tab_completion:
+                            def complete(text,state):
+                                results = [peer for peer in peers_dict if peer.startswith(text)] + [None]
+                                return results[state]
+                            readline.set_completer(complete)                        
+                        
+                        the_client = input('\nPlease type the client name (or regex) you would like to blacklist (and disconnect),'
+                                            ' or the client IP address/subnet you would like to blacklist, or q to quit\n')
+                        if running_with_tab_completion:
+                            readline.set_completer(None) 
+                            
+                        if the_client != 'q':                        
+                            res = input('\nAre you sure you want to blacklist '+the_client+' ? y/n\n')
+                            if res =='y':
+                                #As opposed to ip, the id follows SOURCE_ID_REGEX, which excludes dots
+                                if REGEX_IP.match(the_client):
+                                    task = loop.create_task(connector_remote_tool.add_blacklist_client(client_ip=the_client))
+                                else:
+                                    task = loop.create_task(connector_remote_tool.add_blacklist_client(client_id=the_client))
+                                loop.run_until_complete(task)
+                                print(task.result().decode())                          
+                    else:
+                        print('A client cannot use this functionality')         
+
+                elif the_cmd == 'remove_blacklist_client':
+                    if is_server:
+                        show_attribute('blacklisted_clients_id')
+                        show_attribute('blacklisted_clients_ip')
+                        show_attribute('blacklisted_clients_subnet')                                                
+                        peers_dict = show_connected_peers(return_peers=running_with_tab_completion)
+                        if running_with_tab_completion:
+                            def complete(text,state):
+                                results = [peer for peer in peers_dict if peer.startswith(text)] + [None]
+                                return results[state]
+                            readline.set_completer(complete)                        
+                        
+                        the_client = input('\nPlease type the client name (or regex) you would like to remove from blacklist,'
+                                            ' or the client IP address/subnet you would like to remove from blacklist, or q to quit\n')
+                        if running_with_tab_completion:
+                            readline.set_completer(None) 
+                            
+                        if the_client != 'q':                        
+                            res = input('\nAre you sure you want to remove from blacklist '+the_client+' ? y/n\n')
+                            if res =='y':
+                                if REGEX_IP.match(the_client):
+                                    task = loop.create_task(connector_remote_tool.remove_blacklist_client(client_ip=the_client))
+                                else:
+                                    task = loop.create_task(connector_remote_tool.remove_blacklist_client(client_id=the_client))
+                                loop.run_until_complete(task)
+                                print(task.result().decode())                          
+                    else:
+                        print('A client cannot use this functionality')         
+
+                elif the_cmd == 'add_whitelist_client':
+                    if is_server:
+                        show_attribute('whitelisted_clients_id')
+                        show_attribute('whitelisted_clients_ip')
+                        show_attribute('whitelisted_clients_subnet')                        
+                        
+                        peers_dict = show_connected_peers(return_peers=running_with_tab_completion)
+                        if running_with_tab_completion:
+                            def complete(text,state):
+                                results = [peer for peer in peers_dict if peer.startswith(text)] + [None]
+                                return results[state]
+                            readline.set_completer(complete)                        
+                        
+                        the_client = input('\nPlease type the client name (or regex) you would like to whitelist,'
+                                            ' or the client IP address/subnet you would like to whitelist, or q to quit\n')
+                        if running_with_tab_completion:
+                            readline.set_completer(None) 
+                            
+                        if the_client != 'q':                        
+                            res = input('\nAre you sure you want to whitelist '+the_client+' ? y/n\n')
+                            if res =='y':
+                                if REGEX_IP.match(the_client):
+                                    task = loop.create_task(connector_remote_tool.add_whitelist_client(client_ip=the_client))
+                                else:
+                                    task = loop.create_task(connector_remote_tool.add_whitelist_client(client_id=the_client))
+                                loop.run_until_complete(task)
+                                print(task.result().decode())                          
+                    else:
+                        print('A client cannot use this functionality')         
+
+                elif the_cmd == 'remove_whitelist_client':
+                    if is_server:
+                        show_attribute('whitelisted_clients_id')
+                        show_attribute('whitelisted_clients_ip')
+                        show_attribute('whitelisted_clients_subnet')                        
+                        
+                        peers_dict = show_connected_peers(return_peers=running_with_tab_completion)
+                        if running_with_tab_completion:
+                            def complete(text,state):
+                                results = [peer for peer in peers_dict if peer.startswith(text)] + [None]
+                                return results[state]
+                            readline.set_completer(complete)                        
+                        
+                        the_client = input('\nPlease type the client name (or regex) you would like to remove from whitelist,'
+                                            ' or the client IP address/subnet you would like to remove from whitelist, or q to quit\n')
+                        if running_with_tab_completion:
+                            readline.set_completer(None) 
+                            
+                        if the_client != 'q':                        
+                            res = input('\nAre you sure you want to remove from whitelist '+the_client+' ? y/n\n')
+                            if res =='y':
+                                if REGEX_IP.match(the_client):
+                                    task = loop.create_task(connector_remote_tool.remove_whitelist_client(client_ip=the_client))
+                                else:
+                                    task = loop.create_task(connector_remote_tool.remove_whitelist_client(client_id=the_client))
+                                loop.run_until_complete(task)
+                                print(task.result().decode())                          
+                    else:
+                        print('A client cannot use this functionality')         
+                        
                 elif the_cmd == 'show_log_level':
                     task = loop.create_task(connector_remote_tool.show_log_level())
                     loop.run_until_complete(task)
-                    print(task.result().decode())                                   
+                    print(task.result().decode())      
+                             
                 elif the_cmd == 'set_log_level':
                     list_levels = ['ERROR', 'WARNING', 'INFO', 'DEBUG']
                     dict_levels = {str(index):level for index,level in enumerate(list_levels)}
@@ -217,6 +431,7 @@ def cli(logger=None):
                     task = loop.create_task(connector_remote_tool.set_log_level(new_level))
                     loop.run_until_complete(task)
                     print(task.result().decode())       
+                    
                 elif the_cmd == 'show_subscribe_message_types':
                     if is_server:
                         print('Only available for clients')
@@ -224,6 +439,7 @@ def cli(logger=None):
                         task = loop.create_task(connector_remote_tool.show_subscribe_message_types())
                         loop.run_until_complete(task)
                         print(task.result().decode())     
+                        
                 elif the_cmd == 'set_subscribe_message_types':
                     if is_server:
                         print('Only available for clients')
@@ -384,15 +600,42 @@ def chat(args, logger=None):
     #chat supports sending messages and files/directories between 2 connectors
     if not logger:
         logger = aioconnectors.get_logger(logger_name='chat', first_run=True)
+
+    if not args.nowrap and not args.upload:        
+        try:    
+            import readline
+            readline.set_completer_delims('\t\n= ')            
+            readline.parse_and_bind('tab:complete')
+        except Exception:
+            logger.info('Running without tab completion')
+        else:    
+            proc = subprocess.Popen(
+                    [sys.executable, '-m', 'aioconnectors'] + sys.argv[1:] + ['--nowrap'],
+                    bufsize=0,
+                    stdin=subprocess.PIPE, stdout=None, stderr=None)
+            while True:
+                try:
+                    user_input = input('')                    
+                    proc.stdin.write((user_input + os.linesep).encode())
+                    if user_input == '!exit':
+                        return 
+                except KeyboardInterrupt:
+                    proc.kill()
+                    return
+        
     custom_prompt = 'aioconnectors>> '        
     chat_client_name = 'chat_client'
     CONNECTOR_FILES_DIRPATH = aioconnectors.get_tmp_dir()
+    if os.path.exists(CONNECTOR_FILES_DIRPATH):
+        res = input(f'May I delete the content of {CONNECTOR_FILES_DIRPATH} ? y/n\n')
+        if res == 'y':
+            shutil.rmtree(CONNECTOR_FILES_DIRPATH)
+    
     delete_connector_dirpath_later = not os.path.exists(CONNECTOR_FILES_DIRPATH)
     is_server = not args.target
     accept_all_clients = args.accept
     loop = asyncio.get_event_loop()
     cwd = os.getcwd()
-    
         
     class AuthClient:
         #helper for client authentication on server connector
