@@ -4,8 +4,91 @@ import json
 import subprocess
 import uuid
 import shutil
+import re
 
-from .helpers import get_tmp_dir
+from .helpers import get_tmp_dir, validate_source_id, SOURCE_ID_DEFAULT_REGEX
+
+CA_CREATE_CNF ='''
+[ ca ]
+default_ca    = CA_default      # The default ca section
+
+[ CA_default ]
+
+default_days     = 3650         # How long to certify for
+default_crl_days = 3650         # How long before next CRL
+default_md       = sha256       # Use public key default MD
+preserve         = no           # Keep passed DN ordering
+
+x509_extensions = ca_extensions # The extensions to add to the cert
+
+email_in_dn     = no            # Don't concat the email in the DN
+copy_extensions = copy          # Required to copy SANs from CSR to cert
+
+base_dir      = {base_dir}
+certificate   = $base_dir/{ca_pem}   # The CA certificate
+private_key   = $base_dir/{ca_key}    # The CA private key
+new_certs_dir = $base_dir              # Location for new certs after signing
+database      = $base_dir/index.txt    # Database index file
+serial        = $base_dir/serial.txt   # The current serial number
+
+unique_subject = no  # Set to 'no' to allow creation of
+                     # several certificates with same subject.
+
+[ req ]
+prompt             = no
+default_bits       = 4096
+default_keyfile    = server_ca.pem
+distinguished_name = ca_distinguished_name
+x509_extensions    = ca_extensions
+string_mask        = utf8only
+
+[ ca_distinguished_name ]
+
+countryName                    = US
+#countryName_default            = US
+
+
+[ ca_extensions ]
+
+subjectKeyIdentifier   = hash
+authorityKeyIdentifier = keyid:always, issuer
+basicConstraints       = critical, CA:true
+keyUsage               = keyCertSign, cRLSign
+
+[ signing_policy ]
+countryName            = optional
+stateOrProvinceName    = optional
+localityName           = optional
+organizationName       = supplied
+organizationalUnitName = optional
+commonName             = optional
+emailAddress           = optional
+
+[ signing_req ]
+subjectKeyIdentifier   = hash
+authorityKeyIdentifier = keyid,issuer
+basicConstraints       = CA:FALSE
+keyUsage               = digitalSignature, keyEncipherment
+'''
+
+CA_CSR_CREATE_CNF = '''
+[ req ]
+prompt             = no
+default_bits       = 2048
+default_keyfile    = serverkey.pem
+distinguished_name = server_distinguished_name
+req_extensions     = server_req_extensions
+string_mask        = utf8only
+
+[ server_distinguished_name ]
+countryName                 = US
+organizationName               = {org}
+
+[ server_req_extensions ]
+subjectKeyIdentifier = hash
+basicConstraints     = CA:FALSE
+keyUsage             = digitalSignature, keyEncipherment
+'''
 
 def update_conf(conf_template, conf, replacement_dict):
     shutil.copy(conf_template, conf)
@@ -20,10 +103,10 @@ class SSL_helper:
     CERT_NAME_EXTENSION = "pem"
     KEY_NAME_EXTENSION = "key"
     
-    def __init__(self, logger, is_server, certificates_directory_path=None, max_certs=None):
+    def __init__(self, logger, is_server, certificates_directory_path=None, max_certs=None, server_ca=False):
         self.logger = logger.getChild('ssl')
         try:
-            self.is_server, self.certificates_directory_path = is_server, certificates_directory_path
+            self.is_server, self.certificates_directory_path, self.server_ca = is_server, certificates_directory_path, server_ca
             if not self.certificates_directory_path:
                 self.certificates_directory_path = self.DEFAULT_BASE_PATH
             self.BASE_PATH = self.certificates_directory_path
@@ -39,14 +122,20 @@ class SSL_helper:
             self.SERVER_PEM_PATH = os.path.join(self.SERVER_BASE_PATH, 'server-cert/server.'+self.CERT_NAME_EXTENSION)
             self.SERVER_KEY_PATH = os.path.join(self.SERVER_BASE_PATH, 'server-cert/server.'+self.KEY_NAME_EXTENSION)
             self.CSR_CONF  = os.path.join(self.SERVER_BASE_PATH, 'csr_details.conf')
+            self.SERVER_CA_DETAILS_CONF = os.path.join(self.SERVER_BASE_PATH, 'server_ca_details.conf')            
             self.CSR_TEMPLATE_CONF  = os.path.join(self.SERVER_BASE_PATH, 'csr_details_template.conf')
+            self.SERVER_CA_PEM_PATH = os.path.join(self.SERVER_BASE_PATH, 'server-cert/server_ca.'+self.CERT_NAME_EXTENSION)
+            self.SERVER_CA_KEY_PATH = os.path.join(self.SERVER_BASE_PATH, 'server-cert/server_ca.'+self.KEY_NAME_EXTENSION)
+            self.SERVER_CA_CSR_CONF  = os.path.join(self.SERVER_BASE_PATH, 'server_ca_csr_details.conf')
+            self.SERVER_CA_CSR_PEM_PATH = os.path.join(self.SERVER_BASE_PATH, 'server-cert/server_ca_csr.'+self.CERT_NAME_EXTENSION)
+
             
             #client
             self.CLIENT_BASE_PATH = os.path.join(self.certificates_base_path, 'client')
             self.CLIENT_PEM_PATH  = os.path.join(self.CLIENT_BASE_PATH, 'client-certs/{}.'+self.CERT_NAME_EXTENSION)
             self.CLIENT_KEY_PATH  = os.path.join(self.CLIENT_BASE_PATH, 'client-certs/{}.'+self.KEY_NAME_EXTENSION)
             #self.CLIENT_DEFAULT_ORGANIZATION = '9d2f849c877b4e50b6fccb54d6cd1818'    #'Internet Widgits Pty Ltd'  #'company' 
-            self.CLIENT_SERVER_CRT_PATH  = os.path.join(self.CLIENT_BASE_PATH, 'server-cert/server.crt')                        
+            self.CLIENT_SERVER_CRT_PATH  = os.path.join(self.CLIENT_BASE_PATH, 'server-cert/server.'+self.CERT_NAME_EXTENSION)                        
             #we might want to chain multiple certificates in CLIENT_SERVER_CRT_PATH, to support multiple server certificates
             
             if self.is_server:
@@ -57,12 +146,25 @@ class SSL_helper:
                         self.source_id_2_cert = json.load(fd)
                 else:
                     self.source_id_2_cert = {'source_id_2_cert':{}, 'cert_2_source_id':{}}
-                #load default_client_serial
-                stdout = subprocess.check_output('openssl x509 -hash -serial -noout -in '+\
-                                        str(os.path.join(self.SERVER_CERTS_PATH,
-                                            self.CLIENT_DEFAULT_CERT_NAME+'.'+self.CERT_NAME_EXTENSION)), shell=True)
-                hash_name, serial = stdout.decode().splitlines()
-                self.default_client_serial = serial.split('=')[1]
+                #load default_client_serial, and default_client_serials_list
+                self.default_client_serials_list = []
+                for cert in (file_name for file_name in os.listdir(self.SERVER_CERTS_PATH) if \
+                                                     file_name.endswith(f'.{self.CERT_NAME_EXTENSION}')):
+                    if cert.startswith(self.CLIENT_DEFAULT_CERT_NAME):
+                        cert_name = cert[:-1-len(self.CERT_NAME_EXTENSION)]
+                        if SOURCE_ID_DEFAULT_REGEX.match(cert_name):
+                            stdout = subprocess.check_output('openssl x509 -hash -serial -noout -in '+\
+                                                    str(os.path.join(self.SERVER_CERTS_PATH,
+#                                                        self.CLIENT_DEFAULT_CERT_NAME+'.'+self.CERT_NAME_EXTENSION)), shell=True)
+                                                        cert)), shell=True)
+                            
+                            hash_name, serial = stdout.decode().splitlines()
+                            serial = serial.split('=')[1]
+                            if cert_name == self.CLIENT_DEFAULT_CERT_NAME:
+                                self.default_client_serial = serial
+                            self.logger.info(f'Server adding default certificate : {cert_name}')
+                            self.default_client_serials_list.append(serial)
+                self.logger.info(f'Server using default_client_serials_list : {self.default_client_serials_list}')                 
                 self.max_certs = max_certs
 
         except Exception:
@@ -87,11 +189,13 @@ class SSL_helper:
             certificate = x509.load_pem_x509_certificate(content, default_backend())
     '''
     
-    async def create_client_certificate(self, source_id=None, common_name=None, hook_allow_certificate_creation=None):
+    async def create_client_certificate(self, source_id=None, common_name=None, hook_allow_certificate_creation=None,
+                                        server_ca=False):
         #Only called by server
         #Generates self signed certificate for a client_id
         #returns paths of cert and key
         try:
+            validate_source_id(source_id)
             crt_path = f'{self.SERVER_CERTS_PATH}/{source_id}.{self.CERT_NAME_EXTENSION}'    
             key_path = f'{self.SERVER_CERTS_PATH}/{source_id}.{self.KEY_NAME_EXTENSION}'
             if os.path.exists(crt_path):
@@ -114,14 +218,54 @@ class SSL_helper:
                 #necessary to set a unique field (like organization), so that each certificate has a unique hash, 
                 #which is better for fast authentication
                 organization = uuid.uuid4().hex
-                update_conf(self.CSR_TEMPLATE_CONF, self.CSR_CONF, {'O':organization})                
-                create_certificate_cmd = f"openssl req -new -newkey rsa -nodes -x509 -days 3650 -keyout "\
+                update_conf(self.CSR_TEMPLATE_CONF, self.CSR_CONF, {'O':organization})
+                if not server_ca:
+                    create_certificate_cmd = f"openssl req -new -newkey rsa -nodes -x509 -days 3650 -keyout "\
                                          f"{key_path} -out {crt_path} -config {self.CSR_CONF}"
+                    proc, stdout, stderr = await self.run_cmd(create_certificate_cmd)                
+                    if proc.returncode != 0:
+                        raise Exception('Error while Generating self signed certificate : '+stderr.decode())
+                                         
+                else:
+                    #create csr and sign it with server_ca
+                    with open(self.SERVER_CA_CSR_CONF, 'w') as fd:
+                        fd.write(CA_CSR_CREATE_CNF.format(org=organization))
+                    #update_conf(SERVER_CA_CSR_CONF, csr_details_conf, {'O':organization})            
+                    csr_path = f'{self.SERVER_CERTS_PATH}/{source_id}-csr.{self.CERT_NAME_EXTENSION}'    
+                    
+                    create_csr_cmd = f"openssl req -config {self.SERVER_CA_CSR_CONF} -newkey rsa:2048 -sha256 -nodes -keyout "\
+                                    f"{key_path} -out {csr_path} -outform PEM"
+                    proc, stdout, stderr = await self.run_cmd(create_csr_cmd)                
+                    if proc.returncode != 0:
+                        raise Exception('Error while Generating csr : '+stderr.decode())
+                                    
+                    self.logger.info('Sign client default certificate CSR')
+                    pem_path = f'{self.SERVER_CERTS_PATH}/{source_id}.{self.CERT_NAME_EXTENSION}'    
+                    
+                    # Create the index and serial files
+                    index_file_path = os.path.join(self.SERVER_BASE_PATH, 'index.txt')
+                    serial_file_path = os.path.join(self.SERVER_BASE_PATH, 'serial.txt')
+                    
+                    if not os.path.exists(index_file_path):
+                        open(index_file_path, 'w').close()
+                    if not os.path.exists(serial_file_path):
+                        with open(serial_file_path, 'w') as fd:
+                            fd.write('00')
+                    
+                    create_certificate_cmd = f"openssl ca -batch -policy signing_policy -config {self.SERVER_CA_DETAILS_CONF} "\
+                                            f"-extensions signing_req -out {pem_path} -infiles {csr_path}"
+                    stdout = subprocess.check_output(create_certificate_cmd, shell=True)       
+                    proc, stdout, stderr = await self.run_cmd(create_certificate_cmd)                
+                    if proc.returncode != 0:
+                        raise Exception('Error while Generating CA signed certificate : '+stderr.decode())
+                    
+                    try:
+                        if os.path.exists(csr_path):
+                            #deleting client csr
+                            os.remove(csr_path)
+                    except Exception:
+                        self.logger.exception(f'Deleting {csr_path}')
                                      
-            proc, stdout, stderr = await self.run_cmd(create_certificate_cmd)
-        
-            if proc.returncode != 0:
-                raise Exception('Error while Generating self signed certificate : '+stderr.decode())
             #if stderr:
             #    self.logger.warning('create_certificate_cmd : '+stderr.decode())
         
@@ -134,16 +278,16 @@ class SSL_helper:
             hash_name, serial = stdout.splitlines()
             serial = serial.split('=')[1]
             
-            #create a symlink called <hash>.<first free index, starting from 0>, 
-            #pointing to f'../{source_id}.{self.CERT_NAME_EXTENSION}'
-            index = 0
-            while True:
-                candidate = os.path.join(self.SERVER_SYMLINKS_PATH, hash_name + '.' + str(index))
-                if not os.path.exists(candidate):
-                    break
-                index += 1
-            os.symlink(f'../{source_id}.{self.CERT_NAME_EXTENSION}', candidate)
-
+            if not server_ca:
+                #create a symlink called <hash>.<first free index, starting from 0>, 
+                #pointing to f'../{source_id}.{self.CERT_NAME_EXTENSION}'
+                index = 0
+                while True:
+                    candidate = os.path.join(self.SERVER_SYMLINKS_PATH, hash_name + '.' + str(index))
+                    if not os.path.exists(candidate):
+                        break
+                    index += 1
+                os.symlink(f'../{source_id}.{self.CERT_NAME_EXTENSION}', candidate)
             
             #backup self.source_id_2_cert in file source_id_2_cert.json
             self.source_id_2_cert['source_id_2_cert'][source_id] = serial
@@ -159,6 +303,7 @@ class SSL_helper:
     
     async def remove_client_cert_on_client(self, source_id):
         try:
+            validate_source_id(source_id)
             cert_path = self.CLIENT_PEM_PATH.format(source_id)
             key_path = self.CLIENT_KEY_PATH.format(source_id)
             if not os.path.exists(cert_path):
@@ -180,6 +325,7 @@ class SSL_helper:
         #symlink pointing to f'../{source_id}.{self.CERT_NAME_EXTENSION}' should be removed
         #example : <hash>.2 -> source_id.pem : we need to remove <hash>.2, and then rename all <hash>.i where i>2, to <hash>.i-1 
         try:
+            validate_source_id(source_id)
             cert_path = self.SERVER_CERTS_PEM_PATH.format(source_id)
             key_path = self.SERVER_CERTS_KEY_PATH.format(source_id)
             
@@ -242,7 +388,7 @@ class SSL_helper:
             self.logger.exception('remove_client_cert_on_server')
             return json.dumps({'status':False, 'msg':str(exc)})
 
-def create_certificates(logger, certificates_directory_path):
+def create_certificates(logger, certificates_directory_path, no_ca=False):
     '''
     
     1) Create Server certificate
@@ -260,6 +406,8 @@ def create_certificates(logger, certificates_directory_path):
     4) Create symlink of client default certificate in server directory
     ln -s ../default.pem <hash.0>
     
+    5) Create Server CA certificate
+                                
     '''        
     
     ssl_helper = SSL_helper(logger, is_server=False, certificates_directory_path=certificates_directory_path)
@@ -294,21 +442,33 @@ def create_certificates(logger, certificates_directory_path):
         os.makedirs(certificates_path_client_server)          
     if not os.path.exists(certificates_path_client_client):
         os.makedirs(certificates_path_client_client)                   
-        
-    SERVER_PEM_PATH = os.path.join(certificates_path_server_server, 'server.pem')
-    SERVER_KEY_PATH = os.path.join(certificates_path_server_server, 'server.key')
-    CLIENT_SERVER_CRT_PATH = os.path.join(certificates_path_client_server, 'server.crt')
+
+    CERT_NAME_EXTENSION = "pem"
+    KEY_NAME_EXTENSION = "key"
+
+    #server
+    SERVER_CERTS_PEM_PATH = os.path.join(certificates_path_server_client, '{}.'+CERT_NAME_EXTENSION)
+    SERVER_PEM_PATH = os.path.join(certificates_path_server_server, 'server.'+CERT_NAME_EXTENSION)
+    SERVER_KEY_PATH = os.path.join(certificates_path_server_server, 'server.'+KEY_NAME_EXTENSION)
+    SERVER_CA_PEM = 'server_ca.pem'
+    SERVER_CA_KEY = 'server_ca.key'
+    SERVER_CA_PEM_PATH = os.path.join(certificates_path_server_server, SERVER_CA_PEM)
+    SERVER_CA_KEY_PATH = os.path.join(certificates_path_server_server, SERVER_CA_KEY)
+    SERVER_CA_CSR_CONF  = os.path.join(certificates_path_server, 'server_ca_csr_details.conf')
+    SERVER_CA_CSR_PEM_PATH = os.path.join(certificates_path_server, 'server-cert/server_ca_csr.'+CERT_NAME_EXTENSION)
     
-    CLIENT_PEM_PATH = os.path.join(certificates_path_client_client, '{}.pem')
-    CLIENT_KEY_PATH = os.path.join(certificates_path_client_client, '{}.key')
-    SERVER_CERTS_PATH = os.path.join(certificates_path_server_client, '{}.pem')
+    #client
+    CLIENT_SERVER_CRT_PATH = os.path.join(certificates_path_client_server, 'server.'+CERT_NAME_EXTENSION)
+    CLIENT_PEM_PATH = os.path.join(certificates_path_client_client, '{}.'+CERT_NAME_EXTENSION)
+    CLIENT_KEY_PATH = os.path.join(certificates_path_client_client, '{}.'+KEY_NAME_EXTENSION)
     
-    csr_details_conf = os.path.join(certificates_path_server, 'csr_details.conf')
-    csr_details_template_conf = os.path.join(certificates_path_server, 'csr_details_template.conf')
+    SERVER_CA_DETAILS_CONF = os.path.join(certificates_path_server, 'server_ca_details.conf')    
+    CSR_CONF  = os.path.join(certificates_path_server, 'csr_details.conf')    
+    CSR_TEMPLATE_CONF  = os.path.join(certificates_path_server, 'csr_details_template.conf')    
     
-    #this if lets the user tweak the base csr_details_template_conf
-    if not os.path.exists(csr_details_template_conf):
-        with open(csr_details_template_conf, 'w') as fd:
+    #this if lets the user tweak the base CSR_TEMPLATE_CONF
+    if not os.path.exists(CSR_TEMPLATE_CONF):
+        with open(CSR_TEMPLATE_CONF, 'w') as fd:
             fd.write(
 '''
 [req]
@@ -321,25 +481,76 @@ distinguished_name = dn
 C = US
 ''')
     else:
-        logger.info(f'Using preexisting {csr_details_template_conf}')
-        
+        logger.info(f'Using preexisting {CSR_TEMPLATE_CONF}')
+
+    if not os.path.exists(SERVER_CA_CSR_CONF):
+        with open(SERVER_CA_CSR_CONF, 'w') as fd:
+            fd.write(CA_CSR_CREATE_CNF)
+    else:
+        logger.info(f'Using preexisting {SERVER_CA_CSR_CONF}')
+
+    logger.info('Create Server certificate')
 #        1) Create Server certificate            
-    update_conf(csr_details_template_conf, csr_details_conf, {'O':'company'})                        
-    cmd = f'openssl req -new -newkey rsa -nodes -x509 -days 3650 -keyout {SERVER_KEY_PATH} -out {SERVER_PEM_PATH} -config {csr_details_conf}'
+    update_conf(CSR_TEMPLATE_CONF, CSR_CONF, {'O':'company'})                        
+    cmd = f'openssl req -new -newkey rsa -nodes -x509 -days 3650 -keyout {SERVER_KEY_PATH} -out {SERVER_PEM_PATH} -config {CSR_CONF}'
     stdout = subprocess.check_output(cmd, shell=True)
     shutil.copy(SERVER_PEM_PATH, CLIENT_SERVER_CRT_PATH)
     #we might want to append to an existing CLIENT_SERVER_CRT_PATH, to support multiple server certificates
- 
+
+    if not no_ca:
+        logger.info('Generate CA')
+    #        2)generate ca pem and key
+        if not os.path.exists(SERVER_CA_DETAILS_CONF):
+            with open(SERVER_CA_DETAILS_CONF, 'w') as fd:
+                fd.write(CA_CREATE_CNF.format(base_dir=certificates_path_server_server, ca_pem=SERVER_CA_PEM, ca_key=SERVER_CA_KEY))
+              
+        cmd = f"openssl req -new -newkey rsa:4096 -sha256 -nodes -x509 -days 3650 -keyout {SERVER_CA_KEY_PATH}" \
+              f" -out {SERVER_CA_PEM_PATH} -config {SERVER_CA_DETAILS_CONF} -outform PEM"
+              
+        stdout = subprocess.check_output(cmd, shell=True)        
+        
+        logger.info('Generate client default certificate CSR')
+
 #        2) Create client default certificate
     client_default_key = CLIENT_KEY_PATH.format(ssl_helper.CLIENT_DEFAULT_CERT_NAME)
     client_default_pem = CLIENT_PEM_PATH.format(ssl_helper.CLIENT_DEFAULT_CERT_NAME)
     organization = ssl_helper.CLIENT_DEFAULT_CERT_NAME
     #we might want to obfuscate organization
     organization = str(abs(hash(organization)) % (10 ** 8))
-    update_conf(csr_details_template_conf, csr_details_conf, {'O':organization})                    
-    cmd = f"openssl req -new -newkey rsa -nodes -x509 -days 3650 -keyout {client_default_key} -out {client_default_pem} -config {csr_details_conf}"
-    stdout = subprocess.check_output(cmd, shell=True)        
-    shutil.copy(client_default_pem, SERVER_CERTS_PATH.format(ssl_helper.CLIENT_DEFAULT_CERT_NAME))
+
+    if no_ca:    
+        update_conf(CSR_TEMPLATE_CONF, CSR_CONF, {'O':organization})                    
+        cmd = f"openssl req -new -newkey rsa -nodes -x509 -days 3650 -keyout {client_default_key} -out {client_default_pem} -config {CSR_CONF}"
+        stdout = subprocess.check_output(cmd, shell=True)        
+        shutil.copy(client_default_pem, SERVER_CERTS_PEM_PATH.format(ssl_helper.CLIENT_DEFAULT_CERT_NAME))
+    else:        
+        #create csr and sign it with server_ca
+        if not os.path.exists(SERVER_CA_CSR_CONF):
+            with open(SERVER_CA_CSR_CONF, 'w') as fd:
+                fd.write(CA_CSR_CREATE_CNF.format(org=organization))
+        
+        create_csr_cmd = f"openssl req -config {SERVER_CA_CSR_CONF} -newkey rsa:2048 -sha256 -nodes -keyout "\
+                        f"{client_default_key} -out {SERVER_CA_CSR_PEM_PATH} -outform PEM"
+        stdout = subprocess.check_output(create_csr_cmd, shell=True)              
+    
+        logger.info('Sign client default certificate CSR')
+        
+        # Create the index and serial files
+        index_file_path = os.path.join(certificates_path_server_server, 'index.txt')
+        serial_file_path = os.path.join(certificates_path_server_server, 'serial.txt')
+        
+        if not os.path.exists(index_file_path):
+            open(index_file_path, 'w').close()
+        if not os.path.exists(serial_file_path):
+            with open(serial_file_path, 'w') as fd:
+                fd.write('00')
+        
+        create_certificate_cmd = f"openssl ca -batch -policy signing_policy -config {SERVER_CA_DETAILS_CONF} "\
+                                f"-extensions signing_req -out {client_default_pem} -infiles {SERVER_CA_CSR_PEM_PATH}"
+        stdout = subprocess.check_output(create_certificate_cmd, shell=True)              
+           
+        shutil.copy(client_default_pem, SERVER_CERTS_PEM_PATH.format(ssl_helper.CLIENT_DEFAULT_CERT_NAME))
+
 
 #        3) Calculate hash of client default certificate
     cmd = f'openssl x509 -hash -noout -in {client_default_pem}'
@@ -352,7 +563,8 @@ C = US
         os.remove(dst)
     
     cmd = f'ln -s ../{ssl_helper.CLIENT_DEFAULT_CERT_NAME}.pem '+dst
-    stdout = subprocess.check_output(cmd, shell=True)
+    stdout = subprocess.check_output(cmd, shell=True)    
+    
     logger.info('Finished create_certificates')
     return True
             
